#include "config.h"
#include "defines.h"
#include "mystring.h"
#include "list.h"
#include "hostname.h"
#include "fdbuf.h"
#include "itoa.h"
#include <ctype.h>
#include <errno.h>
#include <getopt.h>
#include <stdlib.h>
#include <string.h>
#include <sys/wait.h>
#include <time.h>
#include <unistd.h>

#define fail(MSG) do{ fout.write("nullmailer-inject: "); fout.writeln(MSG); return false; }while(0)
#define fail_sys(MSG) do{ fout.write("nullmailer-inject: "); fout.write(MSG); fout.write(": "); fout.writeln(strerror(errno)); return false; }while(0)

typedef list<mystring> slist;
static bool do_debug = false;

///////////////////////////////////////////////////////////////////////////////
// Envelope processing
///////////////////////////////////////////////////////////////////////////////
static slist recipients;
static mystring sender;
static bool use_header_recips = true;

mystring parse_address(mystring recip)
{
  recip = recip.strip();
  if(do_debug) {
    fout.write("Parsing address: ");
    fout.write(recip);
  }
  int i;
  int j;
  i = j = -1;			// gets rid of a compiler warning
  if((i = recip.find('<')) >= 0 &&
     (j = recip.find('>', i)) >= i)
    recip = recip.sub(i+1, j-i-1);
  else if((i = recip.find('(')) >= 0 &&
	  (j = recip.find(')', i)) >= i)
    recip = recip.sub(0, i-1).rstrip();
  if(do_debug) {
    fout.write("->");
    fout.writeln(recip);
  }
  return recip;
}

void parse_recip(mystring r)
{
  r = parse_address(r);
  if(!!r)
    recipients.append(r);
}

void parse_recips(const char* str)
{
  const char* start = str;
  const char* end;
  while((end = strchr(start, ',')) != 0) {
    parse_recip(mystring(start, end-start));
    start = end+1;
  }
  parse_recip(start);
}

///////////////////////////////////////////////////////////////////////////////
// Header processing
///////////////////////////////////////////////////////////////////////////////
static slist headers;

static bool header_is_resent = false;

struct header_field
{
  // member information
  const char* name;
  unsigned length;
  bool is_recipient;
  bool is_sender;
  bool is_resent;
  bool remove;

  bool present;

  bool cmp(const mystring& line) 
    {
      if(strncasecmp(line.c_str(), name, length))
	return false;
      if(is_resent) {
	if(!header_is_resent) {
	  sender = "";
	  recipients.empty();
	}
	header_is_resent = true;
      }
      const char* tmp = line.c_str() + length;
      if(is_recipient &&
	 is_resent == header_is_resent &&
	 use_header_recips)
	parse_recips(tmp);
      else if(is_sender && is_resent == header_is_resent && !sender) {
	mystring s = parse_address(tmp);
	if(!!s)
	  sender = s;
      }
      present = true;
      return true;
    }
};

#define F false
#define T true
#define X(N,IR,IS,IRS,R) { #N ":",strlen(#N ":"),\
  IR,IS,IRS,R, false }
static header_field header_fields[] = {
  // Sender address fields, in order of priority
  X(sender,            F,T,F,F),
  X(from,              F,T,F,F),
  X(reply-to,          F,T,F,F),
  X(return-path,       F,T,F,F),
  X(return-receipt-to, F,T,F,F),
  X(errors-to,         F,T,F,F),
  X(resent-sender,     F,T,T,F),
  X(resent-from,       F,T,T,F),
  X(resent-reply-to,   F,T,T,F),
  // Destination address fields
  X(to,                T,F,F,F),
  X(cc,                T,F,F,F),
  X(bcc,               T,F,F,T),
  X(apparently-to,     T,F,F,F),
  X(resent-to,         T,F,T,F),
  X(resent-cc,         T,F,T,F),
  X(resent-bcc,        T,F,T,T),
  // Other fields of interest
  X(date,              F,F,F,F),
  X(message-id,        F,F,F,F),
  X(resent-date,       F,F,T,F),
  X(resent-message-id, F,F,T,F),
};
#define header_field_count (sizeof header_fields/sizeof(header_field))
static bool& header_has_from = header_fields[1].present;
static bool& header_has_to = header_fields[9].present;
static bool& header_has_cc = header_fields[10].present;
static bool& header_has_mid = header_fields[18].present;
static bool& header_has_date = header_fields[17].present;
#undef X
#undef F
#undef T

static mystring from;
static mystring host_name = hostname();

void setup_from()
{
  const char* user = getenv("NULLMAILER_USER");
  if(!user) user = getenv("MAILUSER");
  if(!user) user = getenv("USER");
  if(!user) user = getenv("LOGNAME");
  if(!user) user = "unknown";

  const char* host = getenv("NULLMAILER_HOST");
  if(!host) host = getenv("MAILHOST");
  if(!host) host = getenv("HOSTNAME");
  if(!host) host = host_name.c_str();

  const char* name = getenv("NULLMAILER_NAME");
  if(!name) name = getenv("MAILNAME");
  if(!name) name = getenv("NAME");

  from = mystring(name, " <", user, "@", host, ">");

  const char* suser = getenv("NULLMAILER_SUSER");
  if(!suser) suser = user;

  const char* shost = getenv("NULLMAILER_SHOST");
  if(!shost) shost = host;

  if(!sender && suser && shost)
    sender = mystring(suser, "@", shost);
}

void parse_line(const mystring& line)
{
  bool remove = false;
  for(unsigned i = 0; i < header_field_count; i++) {
    header_field& h = header_fields[i];
    if(h.cmp(line)) {
      remove = h.remove;
      break;
    }
  }
  if(!remove)
    headers.append(line);
}

bool read_header()
{
  mystring line;
  mystring whole;
  while(fin.getline(line)) {
    if(!line)
      break;
    if(isspace(line[0]))
      whole += "\n" + line;
    else {
      if(!!whole)
	parse_line(whole);
      whole = line;
    }
  }
  if(!!whole)
    parse_line(whole);
  return true;
}

mystring make_messageid()
{
  time_t t = time(0);
  char buf[100];
  strftime(buf, 100, "<%Y.%m.%d.%H.%M.%S.", gmtime(&t));
  return mystring(buf, itoa(getpid()), ".nullmailer@", host_name, ">");
}

mystring make_date()
{
  char buf[256];
  time_t t = time(0);
  strftime(buf, 256, "%a, %d %b %Y %H:%M:%S %Z", localtime(&t));
  return buf;
}

bool fix_header()
{
  setup_from();
  if(!header_has_to && !header_has_cc)
    headers.append("Cc: recipient list not shown: ;");
  if(!header_has_date)
    headers.append("Date: " + make_date());
  if(!header_has_from)
    headers.append("From: " + from);
  if(!header_has_mid)
    headers.append("Message-Id: " + make_messageid());
  return true;
}

bool process_header()
{
  return read_header() && fix_header();
}

///////////////////////////////////////////////////////////////////////////////
// Message sending
///////////////////////////////////////////////////////////////////////////////
static fdobuf* nqpipe = 0;
static pid_t pid = 0;

void exec_queue()
{
  if(chdir(BIN_DIR) == -1) {
    fout.write("Could not change directory to BIN_DIR: ");
    fout.writeln(strerror(errno));
    exit(1);
  }
  else
    execl("nullmailer-queue", "nullmailer-queue", 0);
  fout.write("Could not exec nullmailer-queue: ");
  fout.writeln(strerror(errno));
  exit(1);
}

bool start_queue()
{
  int pipe1[2];
  if(pipe(pipe1) == -1)
    fail_sys("Could not create pipe to nullmailer-queue");
  fout.flush();
  pid = fork();
  if(pid == -1)
    fail_sys("Could not fork");
  if(pid == 0) {
    close(pipe1[1]);
    close(0);
    dup2(pipe1[0], 0);
    exec_queue();
  }
  else {
    close(pipe1[0]);
    nqpipe = new fdobuf(pipe1[1], true);
  }
  return true;
}

bool send_env()
{
  if(!nqpipe->write(sender) || !nqpipe->write("\n"))
    fail("Error sending sender to nullmailer-queue.");
  for(slist::iter iter(recipients); iter; iter++)
    if(!nqpipe->write(*iter) || !nqpipe->write("\n"))
      fail("Error sending recipients to nullmailer-queue.");
  if(!nqpipe->writeln(""))
    fail("Error sending recipients to nullmailer-queue.");
  return true;
}

bool send_header()
{
  for(slist::iter iter(headers); iter; iter++)
    if(!nqpipe->write(*iter) || !nqpipe->write("\n"))
      fail("Error sending header to nullmailer-queue.");
  if(!nqpipe->writeln(""))
    fail("Error sending header to nullmailer-queue.");
  return true;
}

bool send_body()
{
  if(!fdbuf_copy(fin, *nqpipe))
    fail("Error sending message body to nullmailer-queue.");
  return true;
}

bool wait_queue()
{
  if(!nqpipe->close())
    fail("Error closing pipe to nullmailer-queue.");
  int status;
  if(waitpid(pid, &status, 0) == -1)
    fail("Error catching the return value from nullmailer-queue.");
  if(WIFEXITED(status)) {
    status = WEXITSTATUS(status);
    if(status)
      fail("nullmailer-queue failed.");
    else
      return true;
  }
  else
    fail("nullmailer-queue crashed or was killed.");
}

bool send_message() 
{
  return start_queue() &&
    send_env() && send_header() && send_body() &&
    wait_queue();
}

///////////////////////////////////////////////////////////////////////////////
// Main
///////////////////////////////////////////////////////////////////////////////
bool parse_args(int argc, char* argv[])
{
  bool result = true;
  for(int i = 1; i < argc; i++) {
    mystring r = parse_address(argv[i]);
    if(!r) {
      fout.write("Invalid recipient: ");
      fout.writeln(argv[i]);
      result = false;
    }
    else {
      recipients.append(r);
      use_header_recips = false;
    }
  }
  return result;
}

int main(int argc, char* argv[])
{
  if(!parse_args(argc, argv))
    return 1;
  process_header() &&
    send_message();
  return 1;
}
